"""JSONalyze Query Engine.

WARNING: This tool executes a SQL prompt generated by the LLM with SQL Lite and
may lead to arbitrary file creation on the machine running this tool.
This tool is not recommended to be used in a production setting, and would
require heavy sandboxing or virtual machines.

DEPRECATED: Use `JSONalyzeQueryEngine` from `llama-index-experimental` instead.

"""

import asyncio
import json
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple

from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.base.response.schema import Response
from llama_index.core.indices.struct_store.sql_retriever import (
    BaseSQLParser,
    DefaultSQLParser,
)
from llama_index.core.llms.llm import LLM
from llama_index.core.prompts import BasePromptTemplate, PromptTemplate
from llama_index.core.prompts.default_prompts import DEFAULT_JSONALYZE_PROMPT
from llama_index.core.prompts.mixin import PromptDictType, PromptMixinType
from llama_index.core.prompts.prompt_type import PromptType
from llama_index.core.schema import QueryBundle
from llama_index.core.settings import Settings
from llama_index.core.utils import print_text

logger = logging.getLogger(__name__)

DEFAULT_RESPONSE_SYNTHESIS_PROMPT_TMPL = (
    "Given a query, synthesize a response based on SQL query results"
    " to satisfy the query. Only include details that are relevant to"
    " the query. If you don't know the answer, then say that.\n"
    "SQL Query: {sql_query}\n"
    "Table Schema: {table_schema}\n"
    "SQL Response: {sql_response}\n"
    "Query: {query_str}\n"
    "Response: "
)

DEFAULT_RESPONSE_SYNTHESIS_PROMPT = PromptTemplate(
    DEFAULT_RESPONSE_SYNTHESIS_PROMPT_TMPL,
    prompt_type=PromptType.SQL_RESPONSE_SYNTHESIS,
)

DEFAULT_TABLE_NAME = "items"


def default_jsonalyzer(
    list_of_dict: List[Dict[str, Any]],
    query_bundle: QueryBundle,
    llm: LLM,
    table_name: str = DEFAULT_TABLE_NAME,
    prompt: BasePromptTemplate = DEFAULT_JSONALYZE_PROMPT,
    sql_parser: BaseSQLParser = DefaultSQLParser(),
) -> Tuple[str, Dict[str, Any], List[Dict[str, Any]]]:
    """Default JSONalyzer that executes a query on a list of dictionaries.

    Args:
        list_of_dict (List[Dict[str, Any]]): List of dictionaries to query.
        query_bundle (QueryBundle): The query bundle.
        llm (LLM): The llm to use.
        table_name (str): The table name to use, defaults to DEFAULT_TABLE_NAME.
        prompt (BasePromptTemplate): The prompt to use.
        sql_parser (BaseSQLParser): The SQL parser to use.

    Returns:
        Tuple[str, Dict[str, Any], List[Dict[str, Any]]]: The SQL Query,
            the Schema, and the Result.
    """
    try:
        import sqlite_utils  # pants: no-infer-dep
    except ImportError as exc:
        IMPORT_ERROR_MSG = (
            "sqlite-utils is needed to use this Query Engine:\n"
            "pip install sqlite-utils"
        )

        raise ImportError(IMPORT_ERROR_MSG) from exc
    # Instantiate in-memory SQLite database
    db = sqlite_utils.Database(memory=True)
    try:
        # Load list of dictionaries into SQLite database
        db[table_name].insert_all(list_of_dict)  # type: ignore
    except sqlite_utils.utils.sqlite3.IntegrityError as exc:
        print_text(f"Error inserting into table {table_name}, expected format:")
        print_text("[{col1: val1, col2: val2, ...}, ...]")
        raise ValueError("Invalid list_of_dict") from exc

    # Get the table schema
    table_schema = db[table_name].columns_dict

    query = query_bundle.query_str
    prompt = prompt or DEFAULT_JSONALYZE_PROMPT
    # Get the SQL query with text-to-SQL prompt
    response_str = llm.predict(
        prompt=prompt,
        table_name=table_name,
        table_schema=table_schema,
        question=query,
    )

    sql_parser = sql_parser or DefaultSQLParser()

    sql_query = sql_parser.parse_response_to_sql(response_str, query_bundle)

    try:
        # Execute the SQL query
        results = list(db.query(sql_query))
    except sqlite_utils.utils.sqlite3.OperationalError as exc:
        print_text(f"Error executing query: {sql_query}")
        raise ValueError("Invalid query") from exc

    return sql_query, table_schema, results


async def async_default_jsonalyzer(
    list_of_dict: List[Dict[str, Any]],
    query_bundle: QueryBundle,
    llm: LLM,
    prompt: Optional[BasePromptTemplate] = None,
    sql_parser: Optional[BaseSQLParser] = None,
    table_name: str = DEFAULT_TABLE_NAME,
) -> Tuple[str, Dict[str, Any], List[Dict[str, Any]]]:
    """Default JSONalyzer.

    Args:
        list_of_dict (List[Dict[str, Any]]): List of dictionaries to query.
        query_bundle (QueryBundle): The query bundle.
        llm (LLM): The llm to use.
        prompt (BasePromptTemplate, optional): The prompt to use.
        sql_parser (BaseSQLParser, optional): The SQL parser to use.
        table_name (str, optional): The table name to use, defaults to DEFAULT_TABLE_NAME.

    Returns:
        Tuple[str, Dict[str, Any], List[Dict[str, Any]]]: The SQL Query,
            the Schema, and the Result.
    """
    try:
        import sqlite_utils  # pants: no-infer-dep
    except ImportError as exc:
        IMPORT_ERROR_MSG = (
            "sqlite-utils is needed to use this Query Engine:\n"
            "pip install sqlite-utils"
        )

        raise ImportError(IMPORT_ERROR_MSG) from exc
    # Instantiate in-memory SQLite database
    db = sqlite_utils.Database(memory=True)
    try:
        # Load list of dictionaries into SQLite database
        db[table_name].insert_all(list_of_dict)  # type: ignore
    except sqlite_utils.utils.sqlite3.IntegrityError as exc:
        print_text(f"Error inserting into table {table_name}, expected format:")
        print_text("[{col1: val1, col2: val2, ...}, ...]")
        raise ValueError("Invalid list_of_dict") from exc

    # Get the table schema
    table_schema = db[table_name].columns_dict

    query = query_bundle.query_str
    prompt = prompt or DEFAULT_JSONALYZE_PROMPT
    # Get the SQL query with text-to-SQL prompt
    response_str = await llm.apredict(
        prompt=prompt,
        table_name=table_name,
        table_schema=table_schema,
        question=query,
    )

    sql_parser = sql_parser or DefaultSQLParser()

    sql_query = sql_parser.parse_response_to_sql(response_str, query_bundle)

    try:
        # Execute the SQL query
        results = list(db.query(sql_query))
    except sqlite_utils.utils.sqlite3.OperationalError as exc:
        print_text(f"Error executing query: {sql_query}")
        raise ValueError("Invalid query") from exc

    return sql_query, table_schema, results


def load_jsonalyzer(
    use_async: bool = False,
    custom_jsonalyzer: Optional[Callable] = None,
) -> Callable:
    """Load the JSONalyzer.

    Args:
        use_async (bool): Whether to use async.
        custom_jsonalyzer (Callable): A custom JSONalyzer to use.

    Returns:
        Callable: The JSONalyzer.
    """
    if custom_jsonalyzer:
        assert not use_async or asyncio.iscoroutinefunction(
            custom_jsonalyzer
        ), "custom_jsonalyzer function must be async when use_async is True"
        return custom_jsonalyzer
    else:
        # make mypy happy to indent this
        if use_async:
            return async_default_jsonalyzer
        else:
            return default_jsonalyzer


class JSONalyzeQueryEngine(BaseQueryEngine):
    """JSON List Shape Data Analysis Query Engine.

    Converts natural language statasical queries to SQL within in-mem SQLite queries.

    list_of_dict(List[Dict[str, Any]]): List of dictionaries to query.
    jsonalyze_prompt (BasePromptTemplate): The JSONalyze prompt to use.
    use_async (bool): Whether to use async.
    analyzer (Callable): The analyzer that executes the query.
    sql_parser (BaseSQLParser): The SQL parser that ensures valid SQL being parsed
        from llm output.
    synthesize_response (bool): Whether to synthesize a response.
    response_synthesis_prompt (BasePromptTemplate): The response synthesis prompt
        to use.
    table_name (str): The table name to use.
    verbose (bool): Whether to print verbose output.
    """

    def __init__(
        self,
        list_of_dict: List[Dict[str, Any]],
        llm: Optional[LLM] = None,
        jsonalyze_prompt: Optional[BasePromptTemplate] = None,
        use_async: bool = False,
        analyzer: Optional[Callable] = None,
        sql_parser: Optional[BaseSQLParser] = None,
        synthesize_response: bool = True,
        response_synthesis_prompt: Optional[BasePromptTemplate] = None,
        table_name: str = DEFAULT_TABLE_NAME,
        verbose: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize params."""
        self._list_of_dict = list_of_dict
        self._llm = llm or Settings.llm
        self._jsonalyze_prompt = jsonalyze_prompt or DEFAULT_JSONALYZE_PROMPT
        self._use_async = use_async
        self._analyzer = load_jsonalyzer(use_async, analyzer)
        self._sql_parser = sql_parser or DefaultSQLParser()
        self._synthesize_response = synthesize_response
        self._response_synthesis_prompt = (
            response_synthesis_prompt or DEFAULT_RESPONSE_SYNTHESIS_PROMPT
        )
        self._table_name = table_name
        self._verbose = verbose

        super().__init__(callback_manager=Settings.callback_manager)

    def _get_prompts(self) -> Dict[str, Any]:
        """Get prompts."""
        return {
            "jsonalyze_prompt": self._jsonalyze_prompt,
            "response_synthesis_prompt": self._response_synthesis_prompt,
        }

    def _update_prompts(self, prompts: PromptDictType) -> None:
        """Update prompts."""
        if "jsonalyze_prompt" in prompts:
            self._jsonalyze_prompt = prompts["jsonalyze_prompt"]
        if "response_synthesis_prompt" in prompts:
            self._response_synthesis_prompt = prompts["response_synthesis_prompt"]

    def _get_prompt_modules(self) -> PromptMixinType:
        """Get prompt sub-modules."""
        return {}

    def _query(self, query_bundle: QueryBundle) -> Response:
        """Answer an analytical query on the JSON List."""
        query = query_bundle.query_str
        if self._verbose:
            print_text(f"Query: {query}\n", color="green")

        # Perform the analysis
        sql_query, table_schema, results = self._analyzer(
            self._list_of_dict,
            query_bundle,
            self._llm,
            table_name=self._table_name,
            prompt=self._jsonalyze_prompt,
            sql_parser=self._sql_parser,
        )
        if self._verbose:
            print_text(f"SQL Query: {sql_query}\n", color="blue")
            print_text(f"Table Schema: {table_schema}\n", color="cyan")
            print_text(f"SQL Response: {results}\n", color="yellow")

        if self._synthesize_response:
            response_str = self._llm.predict(
                self._response_synthesis_prompt,
                sql_query=sql_query,
                table_schema=table_schema,
                sql_response=results,
                query_str=query_bundle.query_str,
            )
            if self._verbose:
                print_text(f"Response: {response_str}", color="magenta")
        else:
            response_str = str(results)
        response_metadata = {"sql_query": sql_query, "table_schema": str(table_schema)}

        return Response(response=response_str, metadata=response_metadata)

    async def _aquery(self, query_bundle: QueryBundle) -> Response:
        """Answer an analytical query on the JSON List."""
        query = query_bundle.query_str
        if self._verbose:
            print_text(f"Query: {query}", color="green")

        # Perform the analysis
        sql_query, table_schema, results = self._analyzer(
            self._list_of_dict,
            query,
            self._llm,
            table_name=self._table_name,
            prompt=self._jsonalyze_prompt,
        )
        if self._verbose:
            print_text(f"SQL Query: {sql_query}\n", color="blue")
            print_text(f"Table Schema: {table_schema}\n", color="cyan")
            print_text(f"SQL Response: {results}\n", color="yellow")

        if self._synthesize_response:
            response_str = await self._llm.apredict(
                self._response_synthesis_prompt,
                sql_query=sql_query,
                table_schema=table_schema,
                sql_response=results,
                query_str=query_bundle.query_str,
            )
            if self._verbose:
                print_text(f"Response: {response_str}", color="magenta")
        else:
            response_str = json.dumps(
                {
                    "sql_query": sql_query,
                    "table_schema": table_schema,
                    "sql_response": results,
                }
            )
        response_metadata = {"sql_query": sql_query, "table_schema": str(table_schema)}

        return Response(response=response_str, metadata=response_metadata)



from llama_index.experimental.query_engine.pandas.pandas_query_engine import (
    PandasQueryEngine,
)
from llama_index.experimental.param_tuner.base import ParamTuner
from llama_index.experimental.nudge.base import Nudge

__all__ = ["PandasQueryEngine", "ParamTuner", "Nudge"]



from llama_index.experimental.query_engine.pandas.pandas_query_engine import (
    PandasQueryEngine,
)
from llama_index.experimental.query_engine.pandas.output_parser import (
    PandasInstructionParser,
)
from llama_index.experimental.query_engine.jsonalyze.jsonalyze_query_engine import (
    JSONalyzeQueryEngine,
)

__all__ = ["PandasQueryEngine", "PandasInstructionParser", "JSONalyzeQueryEngine"]



from llama_index.core.prompts import PromptTemplate, PromptType

############################################
# Pandas
############################################

DEFAULT_PANDAS_TMPL = (
    "You are working with a pandas dataframe in Python.\n"
    "The name of the dataframe is `df`.\n"
    "This is the result of `print(df.head())`:\n"
    "{df_str}\n\n"
    "Follow these instructions:\n"
    "{instruction_str}\n"
    "Query: {query_str}\n\n"
    "Expression:"
)

DEFAULT_PANDAS_PROMPT = PromptTemplate(
    DEFAULT_PANDAS_TMPL, prompt_type=PromptType.PANDAS
)



import logging
from typing import Dict, Optional

from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.utils import infer_torch_device
from llama_index.finetuning.embeddings.common import EmbeddingQAFinetuneDataset

logger = logging.getLogger(__name__)

NUDGE_IMPORT_ERROR_MSG = "NUDGE is not installed. Please install it with 'pip install nudge-ft' to use this functionality."
NUMPY_IMPORT_ERROR_MSG = "Numpy is not installed. Please install it with 'pip install numpy' to use this functionality."
PYTORCH_IMPORT_ERROR_MSG = "Pytorch is not installed. Please install it with 'pip install torch' to use this functionality."


class Nudge:
    """The algorithm implemented here and the current state of the art is called [NUDGE](https://www.arxiv.org/abs/2409.02343).
    If a validation dataset is provided, the best model is evaluated and saved based on the validation loss at the end of every epoch.

    Args:
        train_dataset (EmbeddingQAFinetuneDataset): Dataset to finetune on.
        embed_model (BaseEmbedding): Embedding model.
        val_dataset (EmbeddingQAFinetuneDataset): Validation dataset.
        use_nudge_n (bool): Whether to use NUDGE-N or NUDGE-M. Defaults to True.
        device (Optional[str]): Device to use. Defaults to None.
    """

    def __init__(
        self,
        embed_model: BaseEmbedding,
        train_dataset: EmbeddingQAFinetuneDataset,
        val_dataset: EmbeddingQAFinetuneDataset,
        use_nudge_n: bool = True,
        device: Optional[str] = None,
    ) -> None:
        """Init params."""
        try:
            from nudge import NUDGEN, NUDGEM
        except ImportError:
            raise ImportError(NUDGE_IMPORT_ERROR_MSG)
        try:
            import torch
        except ImportError:
            raise ImportError(PYTORCH_IMPORT_ERROR_MSG)

        if device is None:
            device = infer_torch_device()
            logger.info(f"Use pytorch device: {device}")
        self._target_device = torch.device(device)

        self.embed_model = embed_model
        self.corpus = train_dataset.corpus
        self.corpus_embeddings = self._get_corpus_embeddings(self.corpus)
        self.train_dataset = self._format_dataset(train_dataset, self.corpus)
        self.val_dataset = self._format_dataset(val_dataset, self.corpus)

        self.nudge = (
            NUDGEN(device=self._target_device)
            if use_nudge_n
            else NUDGEM(device=self._target_device)
        )

    def _format_dataset(
        self, dataset: EmbeddingQAFinetuneDataset, corpus: Dict[str, str]
    ):
        """
        Convert the dataset into NUDGE format.

        Args:
            dataset (EmbeddingQAFinetuneDataset): Dataset to convert.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(NUMPY_IMPORT_ERROR_MSG)

        q_embs = []
        q_ans_indx = []
        corpus_keys = list(corpus.keys())

        for query_id, query in dataset.queries.items():
            query_embedding = self.embed_model.get_query_embedding(query)
            q_embs.append(query_embedding)

            relevant_docs = dataset.relevant_docs[query_id]
            relevant_doc_indices = [corpus_keys.index(doc) for doc in relevant_docs]
            q_ans_indx.append(relevant_doc_indices)

        return {"q_embs": np.array(q_embs), "q_ans_indx": q_ans_indx}

    def _get_corpus_embeddings(self, corpus: Dict[str, str]):
        """Get corpus embeddings."""
        try:
            import numpy as np
        except ImportError:
            raise ImportError(NUMPY_IMPORT_ERROR_MSG)

        text_embeddings = [
            self.embed_model.get_text_embedding(text) for text in corpus.values()
        ]
        return np.array(text_embeddings)

    def finetune(self):
        self.corpus_embeddings = self.nudge.finetune_embeddings(
            embeddings=self.corpus_embeddings,
            train_set=self.train_dataset,
            val_set=self.val_dataset,
            nontrain_embeddings=None,
            val_batch_size=256,
            gamma=None,
        )

    def insert_data_and_finetune(
        self,
        new_train_dataset_batch: EmbeddingQAFinetuneDataset,
        new_val_dataset_batch: Optional[EmbeddingQAFinetuneDataset] = None,
    ):
        """
        Insert data and finetune. This should only be done if the new data you are inserting does not conflict with the already existing data. It's important to not finetune multiple times as this can cause the embeddings to lose semantic meaning since they will become further from the original embeddings.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(NUMPY_IMPORT_ERROR_MSG)

        new_corpus_batch = new_train_dataset_batch.corpus
        # if any of the new ids are already in the existing corpus, raise an error
        if any(id in self.corpus for id in new_corpus_batch):
            raise ValueError(
                f"ID {id} already exists in the existing corpus. New IDs must be unique."
            )

        # get the embeddings for the new corpus
        new_corpus_initial_embeddings_batch = self._get_corpus_embeddings(
            new_corpus_batch
        )

        existing_corpus_embeddings = self.corpus_embeddings

        new_train_dataset = self._format_dataset(
            new_train_dataset_batch, new_corpus_batch
        )
        new_val_dataset = self._format_dataset(new_val_dataset_batch, new_corpus_batch)

        new_corpus_embeddings_batch = self.nudge.finetune_embeddings(
            embeddings=new_corpus_initial_embeddings_batch,
            train_set=new_train_dataset,
            val_set=new_val_dataset,
            # runs faster by filtering the embeddings which will not have any queries
            nontrain_embeddings=existing_corpus_embeddings,
            val_batch_size=256,
            gamma=None,
        )

        self.corpus_embeddings = np.concatenate(
            [existing_corpus_embeddings, new_corpus_embeddings_batch]
        )

    def get_finetuned_corpus_embeddings(self):
        return self.corpus_embeddings
